__all__ = ['Sensor']


#  sensor.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

class Sensor:
    def __init__(self) -> None:
        super().__init__()
